# -*- coding: utf-8 -*-
"""stream plugin"""

__author__  = "Adrien DELLE CAVE"
__license__ = """
    Copyright (C) 2017  fjord-technologies

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA..
"""

import bitmath
import boto3
import botocore
import gc
import logging
import os
import Queue
import resource
import threading
import time
import zlib

from dwho.classes.plugins import DWhoPluginBase, PLUGINS
from dwho.config import DWHO_SHARED
from sonicprobe import helpers
from sonicprobe.libs import keystore, threading_tcp_server, workerpool, xbstream
from SocketServer import StreamRequestHandler


LOG = logging.getLogger('xtrargon.plugins.stream')

logging.getLogger('boto3').setLevel(logging.WARNING)
logging.getLogger('botocore').setLevel(logging.WARNING)
logging.getLogger('requests').setLevel(logging.WARNING)
logging.getLogger('xbstream').setLevel(logging.WARNING)

DEFAULT_TIMEOUT         = 60
DEFAULT_COMPRESS_LEVEL  = 6
DEFAULT_MEMORY_PERCENT  = 70

S3_MIN_CHUNK_SIZE       = (5 * 1024 * 1024)
S3_MAX_CHUNK_SIZE       = (20 * 1024 * 1024)
OBJECT_MAX_LIFETIME     = 216000
WORKER_LIFETIME         = 300
JOB_MAX_QUEUE_SIZE      = 25
RLIMIT_NOFILE           = 65536
_KILLED                 = False
_KEYSTORE               = keystore.Keystore()


class XtrargonWriteFile(object):
    def run(self, key):
        res = _KEYSTORE.acquire(key, exists = True)
        if res is None:
            LOG.debug("object writefile: doesn't exist: %r", key)

        (payload, fo) = (None, None)
        try:
            if not _KEYSTORE.get(key, 'filename'):
                return

            dirname  = _KEYSTORE.get(key, 'dirname')
            filename = _KEYSTORE.get(key, 'filename')
            fullpath = os.path.join(dirname, filename)
            helpers.make_dirs(os.path.dirname(fullpath))

            if not _KEYSTORE.get(key, 'fopen'):
                fd = os.open(fullpath, os.O_RDWR|os.O_CREAT)
                _KEYSTORE.set(key, 'fopen', os.fdopen(fd, 'wb+'))

            fo       = _KEYSTORE.get(key, 'fopen')
            size     = _KEYSTORE.get(key, 'size')
            payload  = _KEYSTORE.get(key, 'payload')

            LOG.info("object writef: %r, size: %r", key, size)

            while payload and not _KILLED:
                fo.write(payload.pop(0)[1])

            _KEYSTORE.set(key, 'payload', payload)

            if _KEYSTORE.get(key, 'last_chunk_type') == xbstream.XB_CHUNK_TYPE_EOF:
                if fo:
                    fo.close()
                _KEYSTORE.delete(key)

            LOG.info("object writef wrote: %r, size: %r", key, size)
        except Exception, e:
            if fo:
                fo.close()
            LOG.exception(e)
        finally:
            _KEYSTORE.try_release(key)
            del payload
            gc.collect()


class XtrargonUploadNormal(object):
    def __init__(self, fsdbucket, fsdclient, timeout = None):
        self.fsdbucket = fsdbucket
        self.fsdclient = fsdclient
        self.timeout   = timeout or DEFAULT_TIMEOUT

    def run(self, key):
        res = _KEYSTORE.acquire(key, exists = True)
        if res is None:
            LOG.debug("object normal: doesn't exist: %r", key)

        (payload, x) = (None, None)

        try:
            payload = _KEYSTORE.get(key, 'payload')
            if not payload or _KEYSTORE.get(key, 'id'):
                return

            size    = _KEYSTORE.get(key, 'size')
            _KEYSTORE.set(key, 'payload', [])
            _KEYSTORE.delete(key)

            LOG.info("put object normal: %r, size: %r", key, size)

            self.fsdclient.put_object(Bucket = self.fsdbucket,
                                      Key    = key,
                                      Body   = ''.join(x[1] for x in payload))

            LOG.info("objet normal transferred: %r, size: %r", key, size)
        except Exception, e:
            LOG.exception(e)
        finally:
            _KEYSTORE.try_release(key)
            del payload, x
            gc.collect()


class XtrargonUploadPart(object):
    def __init__(self, fsdbucket, fsdclient, timeout = None):
        self.fsdbucket = fsdbucket
        self.fsdclient = fsdclient
        self.timeout   = timeout or DEFAULT_TIMEOUT

    def run(self, key):
        res = _KEYSTORE.acquire(key, exists = True)
        if res is None:
            LOG.debug("object part: doesn't exist: %r", key)
            return

        (body, body_size, payload) = ("", 0, "")

        try:
            if _KEYSTORE.get(key, 'size') < S3_MIN_CHUNK_SIZE \
               and _KEYSTORE.get(key, 'last_chunk_type') != xbstream.XB_CHUNK_TYPE_EOF:
                return

            payload    = _KEYSTORE.get(key, 'payload')
            if not payload:
                return

            while body_size < S3_MIN_CHUNK_SIZE and payload:
                (s, b)     = payload.pop(0)
                body_size += s
                body      += b
                (s, b)     = (None, None)

            if body_size < S3_MIN_CHUNK_SIZE \
               and _KEYSTORE.get(key, 'last_chunk_type') != xbstream.XB_CHUNK_TYPE_EOF:
                payload.insert(0, (body_size, body))
                _KEYSTORE.set(key, 'payload', payload)
                return

            part_number = _KEYSTORE.get(key, 'part_number') + 1
            upload_id   = _KEYSTORE.get(key, 'id')
            size        = _KEYSTORE.get(key, 'size')

            (_KEYSTORE.set(key, 'payload', payload)
                      .set(key, 'part_number', part_number)
                      .try_release(key))

            payload     = None

            LOG.info("put object part: %r, part: %r, body_size: %r, size: %r",
                     key,
                     part_number,
                     body_size,
                     size)

            part = self.fsdclient.upload_part(Bucket     = self.fsdbucket,
                                              Key        = key,
                                              Body       = body,
                                              UploadId   = upload_id,
                                              PartNumber = part_number)

            LOG.info("objet part transferred: %r, part: %r, size: %r", key, part_number, body_size)

            if not _KEYSTORE.acquire(key, exists = True):
                LOG.error("object part process: object doesn't exist: %r", key)
                return

            size  = _KEYSTORE.get(key, 'size')
            parts = _KEYSTORE.get(key, 'parts')
            parts.append({'PartNumber': part_number,
                          'ETag':       part['ETag']})

            (_KEYSTORE.set(key, 'parts', parts)
                      .set(key, 'size', size - body_size))
        except Exception, e:
            LOG.exception(e)
        finally:
            _KEYSTORE.try_release(key)
            del body, payload
            gc.collect()


class XtrargonUppartComplete(object):
    def __init__(self, fsdbucket, fsdclient, timeout = None):
        self.fsdbucket = fsdbucket
        self.fsdclient = fsdclient
        self.timeout   = timeout or DEFAULT_TIMEOUT

    def run(self, key):
        res = _KEYSTORE.acquire(key, exists = True)
        if res is None:
            LOG.debug("object complete: doesn't exist: %r", key)
            return

        try:
            if _KEYSTORE.get(key, 'size') \
               or _KEYSTORE.get(key, 'last_chunk_type') != xbstream.XB_CHUNK_TYPE_EOF:
                return

            upload_id = _KEYSTORE.get(key, 'id')
            size      = _KEYSTORE.get(key, 'size')
            parts     = _KEYSTORE.get(key, 'parts')

            mupload   = {'Parts': []}
            for part in sorted(parts, key = lambda k: k['PartNumber']):
                mupload['Parts'].append(part)

            LOG.info("object part complete: %r, size: %r, parts: %r", key, size, len(parts))

            self.fsdclient.complete_multipart_upload(
                Bucket          = self.fsdbucket,
                Key             = key,
                UploadId        = upload_id,
                MultipartUpload = mupload)

            LOG.info("objet part complete transferred: %r, size: %r, parts: %r", key, size, len(parts))

            _KEYSTORE.delete(key)
        except Exception, e:
            LOG.exception(e)
        finally:
            _KEYSTORE.try_release(key)
            gc.collect()


class XtrargonUploadJob(threading.Thread):
    def __init__(self, queue, uppool, upnorm, uppart, upcomp, writef, timeout = None):
        threading.Thread.__init__(self)

        self.killed = False

        self.queue   = queue
        self.uppool  = uppool
        self.upnorm  = upnorm
        self.uppart  = uppart
        self.upcomp  = upcomp
        self.writef  = writef
        self.timeout = timeout or DEFAULT_TIMEOUT

    def kill(self):
        self.killed = True

    def run(self):
        while not _KILLED and not self.killed:
            if self.queue.empty():
                continue

            key = self.queue.get(True, 0.1)
            res = _KEYSTORE.try_acquire(key, exists = True, timeout = self.timeout)
            if res is None:
                continue
            elif not res:
                self.queue.put(key)
                continue

            try:
                if _KEYSTORE.expired(key, OBJECT_MAX_LIFETIME):
                    LOG.warning("object expired: %r", key)
                    _KEYSTORE.delete(key)
                elif not _KEYSTORE.has_key(key, 'last_chunk_type'):
                    continue
                elif _KEYSTORE.get(key, 'is_ibdata') or _KEYSTORE.get(key, 'writef'):
                    self.uppool.run(self.writef.run, None, 'writef', *[key])
                elif not _KEYSTORE.get(key, 'id'):
                    _KEYSTORE.try_release(key)
                    self.uppool.run(self.upnorm.run, None, "upnorm", *[key])
                elif _KEYSTORE.get(key, 'size') == 0 \
                   and _KEYSTORE.get(key, 'last_chunk_type') == xbstream.XB_CHUNK_TYPE_EOF:
                    _KEYSTORE.try_release(key)
                    self.uppool.run(self.upcomp.run, None, "upcomp", *[key])
                elif _KEYSTORE.get(key, 'last_chunk_type') == xbstream.XB_CHUNK_TYPE_EOF \
                   or _KEYSTORE.get(key, 'size') > S3_MIN_CHUNK_SIZE:
                    _KEYSTORE.try_release(key)
                    self.uppool.run(self.uppart.run, None, "uppart", *[key])
            except Exception, e:
                LOG.exception(e)
            finally:
                _KEYSTORE.try_release(key)
                self.queue.put(key)
                gc.collect()


class XtrargonJobDispatcher(threading.Thread):
    def __init__(self, uppool, fsdbucket, fsdclient, timeout = None):
        threading.Thread.__init__(self)

        self.uppool  = uppool
        self.upnorm  = XtrargonUploadNormal(fsdbucket, fsdclient, timeout)
        self.uppart  = XtrargonUploadPart(fsdbucket, fsdclient, timeout)
        self.upcomp  = XtrargonUppartComplete(fsdbucket, fsdclient, timeout)
        self.writef  = XtrargonWriteFile()
        self.timeout = timeout or DEFAULT_TIMEOUT
        self.keys    = set()
        self.jobs    = Queue.Queue()

    def run(self):
        while not _KILLED:
            keys  = _KEYSTORE.list()
            if not keys:
                while True:
                    if self.jobs.empty():
                        break
                    self.jobs.get().kill()
                gc.collect()
                time.sleep(0.5)
                continue

            nkeys = set()
            pkeys = self.keys.copy()

            for key in keys:
                if key not in pkeys:
                    nkeys.add(key)

            while not _KILLED and nkeys:
                cjob = None

                if not self.jobs.empty():
                    for n in range(self.jobs.qsize()):
                        job = self.jobs.get()
                        if job.queue.qsize() >= JOB_MAX_QUEUE_SIZE:
                            self.jobs.put(job)
                        else:
                            cjob = job
                            break

                if not cjob:
                    cjob = XtrargonUploadJob(Queue.Queue(),
                                             self.uppool,
                                             self.upnorm,
                                             self.uppart,
                                             self.upcomp,
                                             self.writef,
                                             self.timeout)
                    cjob.setName("upjob:%d" % (self.jobs.qsize() + 1))
                    cjob.start()

                xkeys = nkeys.copy()
                for key in xkeys:
                    if cjob.queue.qsize() >= JOB_MAX_QUEUE_SIZE:
                        break
                    cjob.queue.put(key)
                    LOG.debug("%r: added %r", cjob.name, key)
                    self.keys.add(key)
                    nkeys.remove(key)

                self.jobs.put(cjob)


class XtrargonStreamHandler(StreamRequestHandler):
    def setup(self):
        StreamRequestHandler.setup(self)
        self.path           = None
        self.compress_level = False

        if 'compress' in self.server.config \
           and bool(self.server.config['compress'].get('enabled')):
            self.compress_level = int(self.server.config['compress']
                                          .get('level', DEFAULT_COMPRESS_LEVEL))

        self.memory_percent = float(self.server.config.get('memory_percent'))
        if self.memory_percent <= 0:
            self.memory_percent = DEFAULT_MEMORY_PERCENT

        callbacks         = xbstream.XBStreamDefaultCallbacks(read = self.rfile.read,
                                                              tx   = self._tx)
        self.xbread       = xbstream.XBStreamRead(callbacks, set_payload = True)

    def _mk_realdstpath(self, filepath):
        if not self.server.config.get('paths'):
            return filepath

        xpath = None
        for key in sorted(self.server.config['paths'].keys(), key=len, reverse=True):
            path = os.path.commonprefix([filepath, key])
            if path and path in self.server.config['paths']:
                xpath = path
                break

        if not xpath:
            return filepath

        r   = filepath
        cfg = self.server.config['paths'][path]

        if cfg.get('dst') and filepath.startswith(xpath):
            r = os.path.join(cfg['dst'], filepath[len(xpath):])

        if cfg.get('prefix_server_id'):
            r = os.path.join(os.path.sep, self.server.server_id, r.lstrip(os.path.sep))

        return r

    def _tx(self, xbobj):
        name = xbobj.name

        if self.compress_level is not False:
            name = "%s.z" % xbobj.name

        is_ibdata = name.startswith('ibdata')
        body      = ""
        key       = self._mk_realdstpath(os.path.join(self.path, name)).lstrip('/')
        try:
            if not _KEYSTORE.exists(key):
                _KEYSTORE.add(key)

            if not _KEYSTORE.acquire(key, exists = True):
                LOG.error("object doesn't exist: %r", key)
                return

            if not _KEYSTORE.has_key(key, 'last_chunk_type'):
                (_KEYSTORE.set(key, 'payload', [])
                          .set(key, 'size', 0)
                          .set(key, 'id', None)
                          .set(key, 'part_number', 0)
                          .set(key, 'last_chunk_type', None)
                          .set(key, 'is_ibdata', is_ibdata)
                          .set(key, 'writef', False)
                          .set(key, 'dirname', self.path)
                          .set(key, 'filename', name)
                          .set(key, 'parts', []))

            if helpers.ps_info(os.getpid())['memory'] > self.server.config['memory_percent']:
                _KEYSTORE.set(key, 'writef', True)

            _KEYSTORE.set(key, 'last_chunk_type', xbobj.chunk_type)

            if xbobj.payload_size <= 0:
                return

            body    = xbobj.payload
            size    = xbobj.payload_size

            if self.compress_level is not False:
                body = zlib.compress(body, self.compress_level)
                size = len(body)

            _KEYSTORE.get(key, 'payload').append((size, body))

            size   += _KEYSTORE.get(key, 'size')

            _KEYSTORE.set(key, 'size', size)

            if size < S3_MIN_CHUNK_SIZE:
                return

            if not _KEYSTORE.get(key, 'id'):
                m = self.server.fsdclient.create_multipart_upload(
                        Bucket = self.server.fsdbucket,
                        Key    = key)
                _KEYSTORE.set(key, 'id', m['UploadId'])
        except Exception, e:
            LOG.exception(e)
        finally:
            _KEYSTORE.try_release(key)
            del body, xbobj
            gc.collect()

    def handle(self):
        try:
            state = DWHO_SHARED.get('xtrargon', 'state')

            if not state or state.get('status') != 'processing':
                LOG.error("invalid xtrargon status")
                return

            self.path = state['path']

            while self.xbread.is_streaming():
                self.xbread()
        except Exception, e:
           LOG.exception(e)


class XtrargonStreamServer(threading.Thread, threading_tcp_server.KillableThreadingTCPServer):
    def __init__(self, config, server_id, fsdbucket, fsdclient):
        threading.Thread.__init__(self)
        self.server           = threading_tcp_server.KillableThreadingTCPServer(
                                  config,
                                  (config['listen_addr'],
                                   config['listen_port']),
                                  XtrargonStreamHandler,
                                  name = "stream")
        self.server.config    = config
        self.server.server_id = server_id
        self.server.fsdbucket = fsdbucket
        self.server.fsdclient = fsdclient

    def run(self):
        LOG.info("stream server will now serve")
        while not self.server.killed():
            try:
                self.server.serve_until_killed()
            except (socket.error, select.error), why:
                if errno.EINTR == why[0]:
                    LOG.debug("interrupted system call")
                elif errno.EBADF == why[0] and self.server.killed():
                    LOG.debug("server close")
                else:
                    raise
        LOG.info("stream server exiting")


class XtrargonStreamPlugin(DWhoPluginBase):
    PLUGIN_NAME = 'stream'

    def safe_init(self):
        self.stream_server = None
        self.fsdbucket     = None
        self.uppool        = None

        resource.setrlimit(resource.RLIMIT_NOFILE, (RLIMIT_NOFILE, RLIMIT_NOFILE))

        if self.config['plugins']['stream'].get('credentials'):
            cred = helpers.load_yaml_file(self.config['plugins']['stream']['credentials'])
            if not cred:
                raise ValueError("unable to read credentials")

            for k, v in cred['s3'].iteritems():
                os.environ[k.upper()] = v

            if cred['s3'].get('aws_default_bucket'):
                self.fsdbucket = cred['s3']['aws_default_bucket']

        if not self.fsdbucket and os.environ.get('AWS_DEFAULT_BUCKET'):
            self.fsdbucket = os.environ['AWS_DEFAULT_BUCKET']

        if not self.fsdbucket:
            raise ValueError("missing default bucket setting")

        config    = self.config['plugins']['stream']
        cfgclient = None

        if config and 'client' in config:
            cfgclient = botocore.client.Config(**config['client'])

        self.fsdclient = boto3.client('s3', config = cfgclient)

    def at_start(self):
        timeout             = float(self.config['plugins']['stream'].get('lock_timeout'))
        if timeout < 0:
            timeout = DEFAULT_TIMEOUT

        worker_lifetime     = int(self.config['plugins']['stream'].get('worker_lifetime')) or WORKER_LIFETIME

        self.stream_server  = XtrargonStreamServer(self.config['plugins']['stream'],
                                                   self.server_id,
                                                   self.fsdbucket,
                                                   self.fsdclient)
        self.stream_server.setDaemon(True)
        self.stream_server.start()

        self.uppool         = workerpool.WorkerPool(max_workers = self.fsdclient._client_config.max_pool_connections,
                                                    life_time   = worker_lifetime,
                                                    name        = "uppool")

        upjob               = XtrargonJobDispatcher(self.uppool,
                                                    self.fsdbucket,
                                                    self.fsdclient,
                                                    timeout)
        upjob.setDaemon(True)
        upjob.setName("upjob")
        upjob.start()

    def at_stop(self):
        global _KILLED
        _KILLED = True

        if self.stream_server:
            self.stream_server.kill()

        if self.uppool:
            self.uppool.killall(1)


if __name__ != "__main__":
    def _start():
        PLUGINS.register(XtrargonStreamPlugin())
    _start()
